from struct import pack

from jawa.attribute import Attribute


class EnclosingMethodAttribute(Attribute):
    ADDED_IN = '5.0.0'
    MINIMUM_CLASS_VERSION = (49, 0)

    def __init__(self, table, name_index=None):
        super().__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'EnclosingMethod'
            ).index
        )
        self.class_index = None
        self.method_index = None

    def pack(self):
        return pack('>HH', self.class_index, self.method_index)

    def unpack(self, info):
        self.class_index = info.u2()
        self.method_index = info.u2()



"""
Utility methods for handling oddities in character encoding encountered
when parsing and writing JVM ClassFiles or object serialization archives.

.. note::

    http://bugs.python.org/issue2857 was an attempt in 2008 to get support
    for MUTF-8/CESU-8 into the python core.
"""


def decode_modified_utf8(s: bytes) -> str:
    """
    Decodes a bytestring containing modified UTF-8 as defined in section
    4.4.7 of the JVM specification.

    :param s: bytestring to be converted.
    :returns: A unicode representation of the original string.
    """
    s = bytearray(s)
    buff = []
    buffer_append = buff.append
    ix = 0
    while ix < len(s):
        x = s[ix]
        ix += 1

        if x >> 7 == 0:
            # Just an ASCII character, nothing else to do.
            pass
        elif x >> 6 == 6:
            y = s[ix]
            ix += 1
            x = ((x & 0x1F) << 6) + (y & 0x3F)
        elif x >> 4 == 14:
            y, z = s[ix:ix+2]
            ix += 2
            x = ((x & 0xF) << 12) + ((y & 0x3F) << 6) + (z & 0x3F)
        elif x == 0xED:
            v, w, x, y, z = s[ix:ix+6]
            ix += 5
            x = 0x10000 + (
                ((v & 0x0F) << 16) +
                ((w & 0x3F) << 10) +
                ((y & 0x0F) << 6) +
                (z & 0x3F)
            )
        elif x == 0xC0 and s[ix] == 0x80:
            ix += 1
            x = 0
        buffer_append(x)
    return u''.join(chr(b) for b in buff)


def encode_modified_utf8(u: str) -> bytearray:
    """
    Encodes a unicode string as modified UTF-8 as defined in section 4.4.7
    of the JVM specification.

    :param u: unicode string to be converted.
    :returns: A decoded bytearray.
    """
    final_string = bytearray()

    for c in [ord(char) for char in u]:
        if c == 0x00 or (0x80 < c < 0x7FF):
            final_string.extend([
                (0xC0 | (0x1F & (c >> 6))),
                (0x80 | (0x3F & c))]
            )
        elif c < 0x7F:
            final_string.append(c)
        elif 0x800 < c < 0xFFFF:
            final_string.extend([
                (0xE0 | (0x0F & (c >> 12))),
                (0x80 | (0x3F & (c >> 6))),
                (0x80 | (0x3F & c))]
            )

    return final_string



from struct import unpack_from, calcsize


class BufferStreamReader(object):
    """Stream-like reader over a buffer mimicing the JVM spec types.
    """
    def __init__(self, buff, starting_offset=0):
        self.pos = starting_offset
        self.buff = buff

    def u1(self):
        r = unpack_from('B', self.buff, offset=self.pos)
        self.pos += 1
        return r[0]

    def u2(self):
        r = unpack_from('>H', self.buff, offset=self.pos)
        self.pos += 2
        return r[0]

    def u4(self):
        r = unpack_from('>I', self.buff, offset=self.pos)
        self.pos += 4
        return r[0]

    def unpack(self, fmt):
        size = calcsize(fmt)
        r = unpack_from(fmt, self.buff, offset=self.pos)
        self.pos += size
        return r

    def seek(self, pos):
        self.pos = pos

    def read(self, length=None):
        if length is None:
            r = self.buff[self.pos:]
            self.pos = len(self.buff)
            return r

        r = self.buff[self.pos:self.pos+length]
        self.pos += length
        return r



import code

try:
    from IPython.terminal import embed
except ImportError:
    IPYTHON_SHELL_AVAILABLE = False
    embed = None
else:
    IPYTHON_SHELL_AVAILABLE = True


def start_shell(local_ns=None):
    """Create and immediately drop into a Python shell.

    If IPython version 5 or greater is available it will be used instead
    of the built-in python shell.

    :params local_ns: An optional dict containing the global namespace of
                       the newly created shell.
    :type local_ns: dict
    """
    if IPYTHON_SHELL_AVAILABLE:
        # Don't try to stop IPython from displaying its banner, since
        # it's different in every major version
        terminal = embed.InteractiveShellEmbed(user_ns={})
        terminal.mainloop(local_ns=local_ns, global_ns={})
    else:
        code.interact(banner='', local=local_ns)



from typing import Optional, Callable, Iterator, IO, Sequence
from struct import unpack, pack
from itertools import repeat

from jawa.constants import UTF8
from jawa.util.flags import Flags
from jawa.util.descriptor import method_descriptor
from jawa.attribute import AttributeTable
from jawa.attributes.code import CodeAttribute


class Method(object):
    def __init__(self, cf):
        self._cf = cf
        self.access_flags = Flags('>H', {
            'acc_public': 0x0001,
            'acc_private': 0x0002,
            'acc_protected': 0x0004,
            'acc_static': 0x0008,
            'acc_final': 0x0010,
            'acc_synchronized': 0x0020,
            'acc_bridge': 0x0040,
            'acc_varargs': 0x0080,
            'acc_native': 0x0100,
            'acc_abstract': 0x0400,
            'acc_strict': 0x0800,
            'acc_synthetic': 0x1000
        })
        self._name_index = 0
        self._descriptor_index = 0
        self.attributes = AttributeTable(cf)

    @property
    def descriptor(self) -> UTF8:
        return self._cf.constants[self._descriptor_index]

    @property
    def name(self) -> UTF8:
        return self._cf.constants[self._name_index]

    @property
    def returns(self) -> str:
        return method_descriptor(self.descriptor.value).returns

    @property
    def args(self) -> Sequence:
        return method_descriptor(self.descriptor.value).args

    @property
    def code(self) -> CodeAttribute:
        """
        A shortcut for::

            >> method.attributes.find_one(name='Code')
        """
        return self.attributes.find_one(name='Code')

    def __repr__(self):
        return f'<Method(name={self.name})>'

    def unpack(self, source: IO):
        """
        Read the Method from the file-like object `fio`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param source: Any file-like object providing `read()`
        """
        self.access_flags.unpack(source.read(2))
        self._name_index, self._descriptor_index = unpack('>HH', source.read(4))
        self.attributes.unpack(source)

    def pack(self, out: IO):
        """
        Write the Method to the file-like object `out`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when saving a ClassFile.

        :param out: Any file-like object providing `write()`
        """
        out.write(self.access_flags.pack())
        out.write(pack(
            '>HH',
            self._name_index,
            self._descriptor_index
        ))
        self.attributes.pack(out)


class MethodTable(object):
    def __init__(self, cf):
        self._cf = cf
        self._table = []

    def append(self, method: Method):
        self._table.append(method)

    def find_and_remove(self, f: Callable):
        """
        Removes any and all methods for which `f(method)` returns `True`.
        """
        self._table = [fld for fld in self._table if not f(fld)]

    def remove(self, method: Method):
        """
        Removes a `method` from the table by identity.
        """
        self._table = [fld for fld in self._table if fld is not method]

    def create(self, name: str, descriptor: str,
               code: CodeAttribute=None) -> Method:
        """
        Creates a new method from `name` and `descriptor`. If `code` is not
        ``None``, add a `Code` attribute to this method.
        """
        method = Method(self._cf)
        name = self._cf.constants.create_utf8(name)
        descriptor = self._cf.constants.create_utf8(descriptor)
        method._name_index = name.index
        method._descriptor_index = descriptor.index
        method.access_flags.acc_public = True

        if code is not None:
            method.attributes.create(CodeAttribute)

        self.append(method)
        return method

    def __iter__(self):
        for method in self._table:
            yield method

    def unpack(self, source: IO):
        """
        Read the MethodTable from the file-like object `source`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param source: Any file-like object providing `read()`
        """
        method_count = unpack('>H', source.read(2))[0]
        for _ in repeat(None, method_count):
            method = Method(self._cf)
            method.unpack(source)
            self.append(method)

    def pack(self, out: IO):
        """
        Write the MethodTable to the file-like object `out`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when saving a ClassFile.

        :param out: Any file-like object providing `write()`
        """
        out.write(pack('>H', len(self)))
        for method in self._table:
            method.pack(out)

    def find(self, *, name: str=None, args: str=None, returns: str=None,
             f: Callable=None) -> Iterator[Method]:
        """
        Iterates over the methods table, yielding each matching method. Calling
        without any arguments is equivalent to iterating over the table. For
        example, to get all methods that take three integers and return void::

            for method in cf.methods.find(args='III', returns='V'):
                print(method.name.value)

        Or to get all private methods::

            is_private = lambda m: m.access_flags.acc_private
            for method in cf.methods.find(f=is_private):
                print method.name.value

        :param name: The name of the method(s) to find.
        :param args: The arguments descriptor (ex: ``III``)
        :param returns: The returns descriptor (Ex: ``V``)
        :param f: Any callable which takes one argument (the method).
        """
        for method in self._table:
            if name is not None and method.name.value != name:
                continue

            descriptor = method.descriptor.value
            end_para = descriptor.find(')')

            m_args = descriptor[1:end_para]
            if args is not None and args != m_args:
                continue

            m_returns = descriptor[end_para + 1:]
            if returns is not None and returns != m_returns:
                continue

            if f is not None and not f(method):
                continue

            yield method

    def find_one(self, **kwargs) -> Optional[Method]:
        """
        Same as ``find()`` but returns only the first result.
        """
        return next(self.find(**kwargs), None)

    def __len__(self):
        return len(self._table)



from struct import pack
from collections import namedtuple

from jawa.attribute import Attribute


line_number_entry = namedtuple('line_number_entry', 'start_pc line_number')


class LineNumberTableAttribute(Attribute):
    ADDED_IN = '1.0.2'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, name_index=None):
        super(LineNumberTableAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'LineNumberTable'
            ).index
        )
        self.line_no = []

    def unpack(self, info):
        length = info.u2()
        table = info.unpack('>{0}H'.format(length * 2))

        self.line_no = [
            line_number_entry(*x)
            for x in zip(*[iter(table)] * 2)
        ]

    def pack(self):
        return pack(
            '>H{0}H'.format(len(self.line_no) * 2),
            len(self.line_no),
            *sum(self.line_no, ())
        )

    def __repr__(self):
        return '<LineNumberTableAttribute({0!r})>'.format(self.line_no)



"""
Methods for parsing standard JVM type descriptors for fields and methods.
"""
from collections import namedtuple


JVMType = namedtuple('JVMType', [
    'base_type',
    'dimensions',
    'name'
])

MethodDescriptor = namedtuple('MethodDescriptor', [
    'returns',
    'args',
    'returns_descriptor',
    'args_descriptor',
    'descriptor'
])


def method_descriptor(descriptor: str) -> MethodDescriptor:
    """
    Parses a Method descriptor as described in section 4.3.3 of the JVM
    specification.
    """
    end_para = descriptor.find(')')
    returns = descriptor[end_para + 1:]
    args = descriptor[1:end_para]

    return MethodDescriptor(
        parse_descriptor(returns)[0],
        parse_descriptor(args),
        returns,
        args,
        descriptor
    )


def field_descriptor(descriptor: str) -> str:
    """
    Parses a Field descriptor as described in section 4.3.2 of the JVM
    specification.
    """
    return parse_descriptor(descriptor)[0]


# JVM Descriptor "BaseType" characters to their
# full simple type.
_HUMAN_NAMES = {
    'L': 'reference',
    'B': 'byte',
    'C': 'char',
    'D': 'double',
    'F': 'float',
    'I': 'int',
    'J': 'long',
    'S': 'short',
    'Z': 'boolean',
    'V': 'void'
}


def parse_descriptor(descriptor: str) -> list:
    """
    Uses a tiny state machine to parse JVM descriptors. To get useful wrappers
    around the results, use :py:func:`jawa.core.descriptor.method_descriptor`
    or :py:func:`jawa.core.descriptor.field_descriptor`.
    """
    # States:
    #   10 == NORMAL,
    #   20 == OBJECT REFERENCE
    state = 10
    tokens = []
    token = []
    dimensions = 0
    for char in descriptor:
        if state == 10 and char == 'L':
            state = 20
        elif state == 10 and char == '[':
            dimensions += 1
        elif state == 10:
            tokens.append(JVMType(char, dimensions, _HUMAN_NAMES[char]))
            dimensions = 0
        elif state == 20 and char == ';':
            tokens.append(JVMType('L', dimensions, ''.join(token)))
            dimensions = 0
            state = 10
            del token[:]
        elif state == 20:
            token.append(char)
    return tokens



from struct import pack

from jawa.attribute import Attribute


class ExceptionsAttribute(Attribute):
    ADDED_IN = '1.0.2'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, name_index=None):
        super(ExceptionsAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'Exceptions'
            ).index
        )
        self.exceptions = []

    def unpack(self, info):
        length = info.u2()
        self.exceptions = list(info.unpack('>{0}H'.format(length)))

    def pack(self):
        return pack(
            '>H{0}H'.format(
                len(self.exceptions)
            ),
            len(self.exceptions),
            *self.exceptions
        )

    def __repr__(self):
        return '<ExceptionsAttribute({0!r})>'.format(self.exceptions)



"""
Utilities for reading & writing JVM method bytecode.
"""
import json
import enum
import pkgutil
from struct import unpack, pack, Struct
from itertools import repeat
from collections import namedtuple

Operand = namedtuple('Operand', ['op_type', 'value'])
_Instruction = namedtuple('Instruction', [
    'mnemonic',
    'opcode',
    'operands',
    'pos'
])


class Instruction(_Instruction):
    """
    Represents a single JVM instruction, consisting
    of an opcode and its potential operands.
    """
    __slots__ = ()

    def size_on_disk(self, start_pos=0):
        """
        Returns the size of this instruction and its operands when
        packed. `start_pos` is required for the `tableswitch` and
        `lookupswitch` instruction as the padding depends on alignment.
        """
        # All instructions are at least 1 byte (the opcode itself)
        size = 1
        fmts = opcode_table[self.opcode]['operands']

        if self.wide:
            size += 2
            # Special case for iinc which has a 2nd extended operand.
            if self.opcode == 0x84:
                size += 2
        elif fmts:
            # A simple opcode with simple operands.
            for fmt, _ in fmts:
                size += fmt.size
        elif self.opcode == 0xAB:
            # lookupswitch
            padding = 4 - (start_pos + 1) % 4
            padding = padding if padding != 4 else 0
            size += padding
            # default & npairs
            size += 8
            size += len(self.operands[0]) * 8
        elif self.opcode == 0xAA:
            # tableswitch
            raise NotImplementedError()

        return size

    @property
    def wide(self):
        """
        ``True`` if this instruction needs to be prefixed by the WIDE
        opcode.
        """
        if not opcode_table[self.opcode].get('can_be_wide'):
            return False

        if self.operands[0].value >= 255:
            return True

        if self.opcode == 0x84:
            if self.operands[1].value >= 255:
                return True

        return False

    @property
    def name(self):
        """Alias for mnemonic."""
        return self.mnemonic

    @property
    def details(self):
        """Extended opcode information."""
        return opcode_table[self.opcode]

    @classmethod
    def create(cls, mnemonic_or_op, operands=None):
        op = opcode_table[mnemonic_or_op]

        return cls(
            op['mnemonic'],
            op['op'],
            operands or [],
            0
        )

    def __eq__(self, other):
        return other == self.mnemonic or super().__eq__(other)


class OperandTypes(enum.IntEnum):
    """
    Constants used to determine the "type" of operand on an opcode,
    such as a BRANCH [offset] or a LITERAL [value].
    """
    LITERAL = 10
    LOCAL_INDEX = 20
    CONSTANT_INDEX = 30
    BRANCH = 40
    PADDING = 50


class OperandFmts(enum.Enum):
    UBYTE = Struct('>B')
    BYTE = Struct('>b')
    USHORT = Struct('>H')
    SHORT = Struct('>h')
    INTEGER = Struct('>i')


def write_instruction(fout, start_pos, ins):
    """
    Writes a single instruction of `opcode` with `operands` to `fout`.

    :param fout: Any file-like object providing ``write()``.
    :param start_pos: The current position in the stream.
    :param ins: The `Instruction` to write.
    """
    opcode, operands = ins.opcode, ins.operands
    fmt_operands = opcode_table[opcode]['operands']

    if ins.wide:
        # The "WIDE" prefix
        fout.write(pack('>B', 0xC4))
        # The real opcode.
        fout.write(pack('>B', opcode))
        fout.write(pack('>H', operands[0].value))
        if opcode == 0x84:
            fout.write(pack('>h', operands[1].value))
    elif fmt_operands:
        # A normal simple opcode with simple operands.
        fout.write(pack('>B', opcode))
        for i, (fmt, _) in enumerate(fmt_operands):
            fout.write(fmt.pack(operands[i].value))
    elif opcode == 0xAB:
        # Special case for lookupswitch.
        fout.write(pack('>B', opcode))
        # assemble([
        #     ('lookupswitch', {
        #         2: -3,
        #         4: 5
        #     }, <default>)
        # ])
        padding = 4 - (start_pos + 1) % 4
        padding = padding if padding != 4 else 0
        fout.write(pack(f'{padding}x'))
        fout.write(pack('>ii', operands[1].value, len(operands[0])))
        for key in sorted(operands[0].keys()):
            fout.write(pack('>ii', key, operands[0][key]))
    elif opcode == 0xAA:
        # Special case for table switch.
        fout.write(pack('>B', opcode))
        padding = 4 - (start_pos + 1) % 4
        padding = padding if padding != 4 else 0
        fout.write(pack(f'{padding}x'))
        fout.write(pack(
            f'>iii{len(operands) - 3}i',
            # Default branch offset
            operands[0].value,
            operands[1].value,
            operands[2].value,
            *(o.value for o in operands[3:])
        ))
    else:
        # opcode with no operands.
        fout.write(pack('>B', opcode))


def read_instruction(fio, start_pos):
    """
    Reads a single instruction from `fio` and returns it, or ``None`` if
    the stream is empty.

    :param fio: Any file-like object providing ``read()``.
    :param start_pos: The current position in the stream.
    """
    op = fio.read(1)

    if not op:
        return None

    op = ord(op)

    ins = opcode_table[op]
    operands = ins['operands']
    name = ins['mnemonic']

    final_operands = []
    # Most opcodes have simple operands.
    if operands:
        for fmt, type_ in operands:
            final_operands.append(
                Operand(
                    type_,
                    fmt.value.unpack(fio.read(fmt.value.size))[0]
                )
            )
    # Special case for lookupswitch.
    elif op == 0xAB:
        # Get rid of the alignment padding.
        padding = 4 - (start_pos + 1) % 4
        padding = padding if padding != 4 else 0
        fio.read(padding)

        # Default branch address and branch count.
        default, npairs = unpack('>ii', fio.read(8))

        pairs = {}
        for _ in repeat(None, npairs):
            match, offset = unpack('>ii', fio.read(8))
            pairs[match] = offset

        final_operands.append(pairs)
        final_operands.append(Operand(OperandTypes.BRANCH, default))
    # Special case for tableswitch
    elif op == 0xAA:
        # Get rid of the alignment padding.
        padding = 4 - (start_pos + 1) % 4
        padding = padding if padding != 4 else 0
        fio.read(padding)

        default, low, high = unpack('>iii', fio.read(12))
        final_operands.append(Operand(OperandTypes.BRANCH, default))
        final_operands.append(Operand(OperandTypes.LITERAL, low))
        final_operands.append(Operand(OperandTypes.LITERAL, high))

        for _ in repeat(None, high - low + 1):
            offset = unpack('>i', fio.read(4))[0]
            final_operands.append(Operand(OperandTypes.BRANCH, offset))
    # Special case for the wide prefix
    elif op == 0xC4:
        real_op = unpack('>B', fio.read(1))[0]
        ins = opcode_table[real_op]
        name = ins['mnemonic']

        final_operands.append(Operand(
            OperandTypes.LOCAL_INDEX,
            unpack('>H', fio.read(2))[0]
        ))
        # Further special case for iinc.
        if real_op == 0x84:
            final_operands.append(Operand(
                OperandTypes.LITERAL,
                unpack('>H', fio.read(2))[0]
            ))

    return Instruction(name, op, final_operands, start_pos)


def load_bytecode_definitions(*, path=None) -> dict:
    """Load bytecode definitions from JSON file.

    If no path is provided the default bytecode.json will be loaded.

    :param path: Either None or a path to a JSON file to load containing
                 bytecode definitions.
    """
    if path is not None:
        with open(path, 'rb') as file_in:
            j = json.load(file_in)
    else:
        try:
            j = json.loads(pkgutil.get_data('jawa.util', 'bytecode.json'))
        except json.JSONDecodeError:
            # Unfortunately our best way to handle missing/malformed/empty
            # bytecode.json files since it may not actually be backed by a
            # "real" file.
            return {}

    for definition in j.values():
        # If the entry has any operands take the text labels and convert
        # them into pre-cached struct objects and operand types.
        operands = definition['operands']
        if operands:
            definition['operands'] = [
                [getattr(OperandFmts, oo[0]), OperandTypes[oo[1]]]
                for oo in operands
            ]

    # Return one dict that contains both mnemonic keys and opcode keys.
    return {**j, **{v['op']: v for v in j.values()}}


opcode_table = load_bytecode_definitions()



"""
Implemented Attributes
======================

In addition to standard JVM attributes various compilers, debuggers, and other
tools may insert additional unknown attributes into a :class:`~jawa.cf
.ClassFile`.

Not all standard attributes are currently implemented. When an unknown
attribute is encountered, an :class:`~jawa.attribute.UnknownAttribute` object
is created instead. This UnknownAttribute retains the name and content of the
original attribute, allowing you to parse it yourself or to simply pass it
through.
"""


import os
import os.path
from typing import IO, Callable, Iterable, Set, Iterator
from itertools import repeat
from zipfile import ZipFile
from collections import OrderedDict
from contextlib import contextmanager

from jawa.cf import ClassFile
from jawa.constants import ConstantPool, ConstantClass


def _walk(path, follow_links=False, maximum_depth=None):
    """A modified os.walk with support for maximum traversal depth."""
    root_level = path.rstrip(os.path.sep).count(os.path.sep)
    for root, dirs, files in os.walk(path, followlinks=follow_links):
        yield root, dirs, files
        if maximum_depth is None:
            continue

        if root_level + maximum_depth <= root.count(os.path.sep):
            del dirs[:]


class ClassLoader(object):
    """Emulate the Java ClassPath.

    Provides utilities for managing a java classpath as well as loading
    classes from those paths.

    :param sources: Optional sources to pass into update().
    :param max_cache: The maximum number of ClassFile's to store in the cache.
                      If set to 0, the cache will be unlimited. [default: 50]
    :type max_cache: Long
    :param klass: The class to use when constructing ClassFiles.
    :type klass: ClassFile or subclass.
    :param bytecode_transforms: Default transforms to apply when disassembling
                                a method.
    """
    def __init__(self, *sources, max_cache: int=50, klass=ClassFile,
                 bytecode_transforms: Iterable[Callable]=None):
        self.path_map = {}
        self.max_cache = max_cache
        self.class_cache = OrderedDict()
        self.bytecode_transforms = bytecode_transforms or []
        self.klass = klass

        if sources:
            self.update(*sources)

    def __getitem__(self, path: str) -> ClassFile:
        return self.load(path)

    def __contains__(self, path: str) -> bool:
        if path in self.path_map:
            return True
        elif path + '.class' in self.path_map:
            return True
        return False

    def update(self, *sources, follow_symlinks: bool=False,
               maximum_depth: int=20):
        """Add one or more ClassFile sources to the class loader.

        If a given source is a directory path, it is traversed up to the
        maximum set depth and all files under it are added to the class loader
        lookup table.

        If a given source is a .jar or .zip file it will be opened and the
        file index added to the class loader lookup table.

        If a given source is a ClassFile or a subclass, it's immediately
        added to the class loader lookup table and the class cache.

        :param sources: One or more ClassFile sources to be added.
        :param follow_symlinks: True if symlinks should be followed when
                                traversing filesystem directories.
                                [default: False]
        :param maximum_depth: The maximum sub-directory depth when traversing
                              filesystem directories. If set to `None` no limit
                              will be enforced. [default: 20]
        """
        for source in sources:
            if isinstance(source, self.klass):
                self.path_map[source.this.name.value] = source
                self.class_cache[source.this.name.value] = source
                continue

            # Explicit cast to str to support Path objects.
            source = str(source)
            if source.lower().endswith(('.zip', '.jar')):
                zf = ZipFile(source, 'r')
                self.path_map.update(zip(zf.namelist(), repeat(zf)))
            elif os.path.isdir(source):
                walker = _walk(
                    source,
                    follow_links=follow_symlinks,
                    maximum_depth=maximum_depth
                )
                for root, dirs, files in walker:
                    for file_ in files:
                        path_full = os.path.join(root, file_)
                        path_suffix = os.path.relpath(path_full, source)
                        self.path_map[path_suffix] = path_full

    @contextmanager
    def open(self, path: str, mode: str='r') -> IO:
        """Open an IO-like object for `path`.

        .. note::

            Mode *must* be either 'r' or 'w', as the underlying objects
            do not understand the full range of modes.

        :param path: The path to open.
        :param mode: The mode of the file being opened, either 'r' or 'w'.
        """
        entry = self.path_map.get(path)
        if entry is None:
            raise FileNotFoundError()

        if isinstance(entry, str):
            with open(entry, 'rb' if mode == 'r' else mode) as source:
                yield source
        elif isinstance(entry, ZipFile):
            with entry.open(path, mode) as source:
                yield source
        else:
            raise NotImplementedError()

    def load(self, path: str) -> ClassFile:
        """Load the class at `path` and return it.

        Load will attempt to load the file at `path` and `path` + .class
        before failing.

        :param path: Fully-qualified path to a ClassFile.
        """
        # Try to refresh the class from the cache, loading it from disk
        # if not found.
        try:
            r = self.class_cache.pop(path)
        except KeyError:
            with self.open(f'{path}.class') as source:
                r = self.klass(source)

        r.classloader = self
        # Even if it was found re-set the key to update the OrderedDict
        # ordering.
        self.class_cache[path] = r

        # If the cache is enabled remove every item over N started from
        # the least-used.
        if self.max_cache > 0:
            to_pop = max(len(self.class_cache) - self.max_cache, 0)
            for _ in repeat(None, to_pop):
                self.class_cache.popitem(last=False)

        return r

    def clear(self):
        """Erase all stored paths and all cached classes."""
        self.path_map.clear()
        self.class_cache.clear()

    def dependencies(self, path: str) -> Set[str]:
        """Returns a set of all classes referenced by the ClassFile at
        `path` without reading the entire ClassFile.

        This is an optimization method that does not load a complete ClassFile,
        nor does it add the results to the ClassLoader cache.

        :param path: Fully-qualified path to a ClassFile.
        """
        return set(c.name.value for c in self.search_constant_pool(
            path=path,
            type_=ConstantClass
        ))

    def search_constant_pool(self, *, path: str, **options):
        """Partially load the class at `path`, yield all matching constants
        from the ConstantPool.

        This is an optimization method that does not load a complete ClassFile,
        nor does it add the results to the ClassLoader cache.

        :param path: Fully-qualified path to a ClassFile.
        :param options: A list of options to pass into `ConstantPool.find()`
        """
        with self.open(f'{path}.class') as source:
            # Skip over the magic, minor, and major version.
            source.read(8)
            pool = ConstantPool()
            pool.unpack(source)
            yield from pool.find(**options)

    @property
    def classes(self) -> Iterator[str]:
        """Yield the name of all classes discovered in the path map."""
        yield from (
            c[:-6]
            for c in self.path_map.keys() if c.endswith('.class')
        )



"""
Transforms are simple Instruction modifiers that can be called on each
Instruction by the :func:`~jawa.attributes.code.CodeAttribute.disassemble`
function.
"""
from jawa.util.bytecode import Instruction, Operand, OperandTypes, opcode_table


def expand_constants(ins: Instruction, *, cf) -> Instruction:
    """Replace CONSTANT_INDEX operands with the literal Constant object from
    the constant pool.

    :param ins: Instruction to potentially modify.
    :param cf: The ClassFile instance used to resolve Constants.
    :return: Potentially modified instruction.
    """
    for i, operand in enumerate(ins.operands):
        if not isinstance(operand, Operand):
            continue

        if operand.op_type == OperandTypes.CONSTANT_INDEX:
            ins.operands[i] = cf.constants[operand.value]

    return ins


def simple_swap(ins: Instruction) -> Instruction:
    """Replaces one instruction with another based on the transform rules in
    the bytecode definitions. This can help simplify your code as it reduces
    the overall number of instructions. For example, `aload_0` will become
    `aload 0`.

    :param ins: Instruction to potentially modify.
    :return: Potentially modified instruction.
    """
    try:
        rule = ins.details['transform']['simple_swap']
    except KeyError:
        return ins

    replacement_ins = opcode_table[rule['op']]

    return Instruction(
        replacement_ins['mnemonic'],
        replacement_ins['op'],
        [Operand(
            replacement_ins['operands'][i][1],
            r
        ) for i, r in enumerate(rule['operands'])],
        ins.pos
    )



from typing import IO, Callable, Iterator, Optional
from struct import unpack, pack
from itertools import repeat

from jawa.util.flags import Flags
from jawa.attribute import AttributeTable
from jawa.constants import Constant, UTF8
from jawa.attributes.constant_value import ConstantValueAttribute


class Field(object):
    def __init__(self, cf):
        self._cf = cf
        self.access_flags = Flags('>H', {
            'acc_public': 0x0001,
            'acc_private': 0x0002,
            'acc_protected': 0x0004,
            'acc_static': 0x0008,
            'acc_final': 0x0010,
            'acc_volatile': 0x0040,
            'acc_transient': 0x0080,
            'acc_synthetic': 0x1000,
            'acc_enum': 0x4000
        })
        self._name_index = 0
        self._descriptor_index = 0
        self.attributes = AttributeTable(cf)

    @property
    def descriptor(self) -> UTF8:
        return self._cf.constants[self._descriptor_index]

    @property
    def name(self) -> UTF8:
        return self._cf.constants[self._name_index]

    @property
    def value(self) -> ConstantValueAttribute:
        """
        A shortcut for the field's ConstantValue attribute, should one exist.
        """
        return self.attributes.find_one(name='ConstantValue')

    def unpack(self, source: IO):
        """
        Read the Field from the file-like object `fio`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param source: Any file-like object providing `read()`
        """
        self.access_flags.unpack(source.read(2))
        self._name_index, self._descriptor_index = unpack('>HH', source.read(4))
        self.attributes.unpack(source)

    def pack(self, out: IO):
        """
        Write the Field to the file-like object `out`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when saving a ClassFile.

        :param out: Any file-like object providing `write()`
        """
        out.write(self.access_flags.pack())
        out.write(pack('>HH', self._name_index, self._descriptor_index))
        self.attributes.pack(out)


class FieldTable(object):
    def __init__(self, cf):
        self._cf = cf
        self._table = []

    def append(self, field: Field):
        self._table.append(field)

    def find_and_remove(self, f: Callable):
        """
        Removes any and all fields for which `f(field)` returns `True`.
        """
        self._table = [fld for fld in self._table if not f(fld)]

    def remove(self, field: Field):
        """
        Removes a `Field` from the table by identity.
        """
        self._table = [fld for fld in self._table if fld is not field]

    def create(self, name: str, descriptor: str, value: Constant=None) -> Field:
        """
        Creates a new field from `name` and `descriptor`. For example::

            >>> from jawa.cf import ClassFile
            >>> cf = ClassFile.create('BeerCounter')
            >>> field = cf.fields.create('BeerCount', 'I')

        To automatically create a static field, pass a value::

            >>> from jawa.cf import ClassFile
            >>> cf = ClassFile.create('BeerCounter')
            >>> field = cf.fields.create(
            ...     'MaxBeer',
            ...     'I',
            ...     cf.constants.create_integer(99)
            ... )

        :param name: Name of the new field.
        :param descriptor: Type descriptor of the new field.
        :param value: Optional static value for the field.
        """
        field = Field(self._cf)
        name = self._cf.constants.create_utf8(name)
        descriptor = self._cf.constants.create_utf8(descriptor)
        field._name_index = name.index
        field._descriptor_index = descriptor.index
        field.access_flags.acc_public = True

        if value is not None:
            field.attributes.create(ConstantValueAttribute, value)
            field.access_flags.acc_static = True

        self.append(field)
        return field

    def __iter__(self):
        for field in self._table:
            yield field

    def unpack(self, source: IO):
        """
        Read the FieldTable from the file-like object `source`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param source: Any file-like object providing `read()`
        """
        field_count = unpack('>H', source.read(2))[0]
        for _ in repeat(None, field_count):
            field = Field(self._cf)
            field.unpack(source)
            self.append(field)

    def pack(self, out: IO):
        """
        Write the FieldTable to the file-like object `out`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when saving a ClassFile.

        :param out: Any file-like object providing `write()`
        """
        out.write(pack('>H', len(self)))
        for field in self._table:
            field.pack(out)

    def __len__(self):
        return len(self._table)

    def find(self, *, name: str=None, type_: str=None,
             f: Callable=None) -> Iterator[Field]:
        """
        Iterates over the fields table, yielding each matching method. Calling
        without any arguments is equivalent to iterating over the table.

        :param name: The name of the field(s) to find.
        :param type_: The field descriptor (Ex: 'I')
        :param f: Any callable which takes one argument (the field).
        """
        for field in self._table:
            if name is not None and field.name.value != name:
                continue

            descriptor = field.descriptor.value
            if type_ is not None and type_ != descriptor:
                continue

            if f is not None and not f(field):
                continue

            yield field

    def find_one(self, **kwargs) -> Optional[Field]:
        """
        Same as ``find()`` but returns only the first result.
        """
        return next(self.find(**kwargs), None)



"""
ClassFile reader & writer.

The :mod:`jawa.cf` module provides tools for working with JVM ``.class``
ClassFiles.
"""
from typing import IO, Iterable, Union, Sequence
from struct import pack, unpack
from collections import namedtuple

from jawa.constants import ConstantPool, ConstantClass
from jawa.fields import FieldTable
from jawa.methods import MethodTable
from jawa.attribute import AttributeTable, ATTRIBUTE_CLASSES
from jawa.util.flags import Flags
from jawa.attributes.bootstrap import BootstrapMethod


class ClassVersion(namedtuple('ClassVersion', ['major', 'minor'])):
    """ClassFile file format version."""

    __slots__ = ()

    @property
    def human(self) -> str:
        """
        A human-readable string identifying this version.

        If the version is unknown, `None` is returned instead.
        """
        return {
            0x33: 'J2SE_7',
            0x32: 'J2SE_6',
            0x31: 'J2SE_5',
            0x30: 'JDK1_4',
            0x2F: 'JDK1_3',
            0x2E: 'JDK1_2',
            0x2D: 'JDK1_1',
        }.get(self.major, None)


class ClassFile(object):
    """
    Implements the JVM ClassFile (files typically ending in ``.class``).

    To open an existing ClassFile::

        >>> with open('HelloWorld.class', 'rb') as fin:
        ...    cf = ClassFile(fin)

    To save a newly created or modified ClassFile::

        >>> cf = ClassFile.create('HelloWorld')
        >>> with open('HelloWorld.class', 'wb') as out:
        ...    cf.save(out)

    :meth:`~ClassFile.create` sets up some reasonable defaults equivalent to:

    .. code-block:: java

        public class HelloWorld extends java.lang.Object{
        }

    :param source: any file-like object providing ``.read()``.
    """

    #: The JVM ClassFile magic number.
    MAGIC = 0xCAFEBABE

    def __init__(self, source: IO=None):
        # Default to J2SE_7
        self._version = ClassVersion(0x32, 0)
        self._constants = ConstantPool()
        self.access_flags = Flags('>H', {
            'acc_public': 0x0001,
            'acc_final': 0x0010,
            'acc_super': 0x0020,
            'acc_interface': 0x0200,
            'acc_abstract': 0x0400,
            'acc_synthetic': 0x1000,
            'acc_annotation': 0x2000,
            'acc_enum': 0x4000
        })
        self._this = 0
        self._super = 0
        self._interfaces = []
        self.fields = FieldTable(self)
        self.methods = MethodTable(self)
        self.attributes = AttributeTable(self)
        #: The ClassLoader bound to this ClassFile, if any.
        self.classloader = None

        if source:
            self._from_io(source)

    @classmethod
    def create(cls, this: str, super_: str=u'java/lang/Object') -> 'ClassFile':
        """
        A utility which sets up reasonable defaults for a new public class.

        :param this: The name of this class.
        :param super_: The name of this class's superclass.
        """
        cf = ClassFile()
        cf.access_flags.acc_public = True
        cf.access_flags.acc_super = True

        cf.this = cf.constants.create_class(this)
        cf.super_ = cf.constants.create_class(super_)

        return cf

    def save(self, source: IO):
        """
        Saves the class to the file-like object `source`.

        :param source: Any file-like object providing write().
        """
        write = source.write

        write(pack(
            '>IHH',
            ClassFile.MAGIC,
            self.version.minor,
            self.version.major
        ))

        self._constants.pack(source)

        write(self.access_flags.pack())
        write(pack(
            f'>HHH{len(self._interfaces)}H',
            self._this,
            self._super,
            len(self._interfaces),
            *self._interfaces
        ))

        self.fields.pack(source)
        self.methods.pack(source)
        self.attributes.pack(source)

    def _from_io(self, source: IO):
        """
        Loads an existing JVM ClassFile from any file-like object.
        """
        read = source.read

        if unpack('>I', source.read(4))[0] != ClassFile.MAGIC:
            raise ValueError('invalid magic number')

        # The version is swapped on disk to (minor, major), so swap it back.
        self.version = unpack('>HH', source.read(4))[::-1]

        self._constants.unpack(source)

        # ClassFile access_flags, see section #4.1 of the JVM specs.
        self.access_flags.unpack(read(2))

        # The CONSTANT_Class indexes for "this" class and its superclass.
        # Interfaces are a simple list of CONSTANT_Class indexes.
        self._this, self._super, interfaces_count = unpack('>HHH', read(6))
        self._interfaces = unpack(
            f'>{interfaces_count}H',
            read(2 * interfaces_count)
        )

        self.fields.unpack(source)
        self.methods.unpack(source)
        self.attributes.unpack(source)

    @property
    def version(self) -> ClassVersion:
        """
        The :class:`~jawa.cf.ClassVersion` for this class.

        Example::

            >>> cf = ClassFile.create('HelloWorld')
            >>> cf.version = 51, 0
            >>> print(cf.version)
            ClassVersion(major=51, minor=0)
            >>> print(cf.version.major)
            51
        """
        return self._version

    @version.setter
    def version(self, major_minor: Union[ClassVersion, Sequence]):
        self._version = ClassVersion(*major_minor)

    @property
    def constants(self) -> ConstantPool:
        """
        The :class:`~jawa.cp.ConstantPool` for this class.
        """
        return self._constants

    @property
    def this(self) -> ConstantClass:
        """
        The :class:`~jawa.constants.ConstantClass` which represents this class.
        """
        return self.constants.get(self._this)

    @this.setter
    def this(self, value):
        self._this = value.index

    @property
    def super_(self) -> ConstantClass:
        """
        The :class:`~jawa.constants.ConstantClass` which represents this
        class's superclass.
        """
        return self.constants.get(self._super)

    @super_.setter
    def super_(self, value: ConstantClass):
        self._super = value.index

    @property
    def interfaces(self) -> Iterable[ConstantClass]:
        """
        A list of direct superinterfaces of this class as indexes into
        the constant pool, in left-to-right order.
        """
        return [self._constants[idx] for idx in self._interfaces]

    @property
    def bootstrap_methods(self) -> BootstrapMethod:
        """
        Returns the bootstrap methods table from the BootstrapMethods attribute,
        if one exists. If it does not, one will be created.

        :returns: Table of `BootstrapMethod` objects.
        """
        bootstrap = self.attributes.find_one(name='BootstrapMethods')

        if bootstrap is None:
            bootstrap = self.attributes.create(
                ATTRIBUTE_CLASSES['BootstrapMethods']
            )

        return bootstrap.table

    def __repr__(self):
        return f'<ClassFile(this={self.this.name.value!r})>'



from struct import pack
from jawa.attribute import Attribute


class SourceFileAttribute(Attribute):
    ADDED_IN = '1.0.2'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, name_index=None):
        super(SourceFileAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                u'SourceFile'
            ).index
        )
        self.source_file_index = None

    def unpack(self, info):
        self.source_file_index = info.u2()

    def pack(self):
        return pack('>H', self.source_file_index)

    @property
    def source_file(self):
        return self.cf.constants[self.source_file_index]

    @source_file.setter
    def source_file(self, value):
        self.source_file_index = value.index



from itertools import repeat

from jawa.attribute import Attribute
from jawa.util.verifier import VerificationTypes

# These types are followed by an additional u2.
TYPES_WITH_EXTRA = (
    VerificationTypes.ITEM_Object,
    VerificationTypes.ITEM_Uninitialized
)


class StackMapFrame(object):
    __slots__ = (
        'frame_type',
        'frame_offset',
        'frame_locals',
        'frame_stack'
    )

    def __init__(self, frame_type):
        self.frame_type = frame_type
        self.frame_offset = 0
        self.frame_locals = []
        self.frame_stack = []

    def __repr__(self):
        return (
            u'<StackMapFrame(type={s.frame_type!r},'
            u'offset={s.frame_offset!r},'
            u'locals={s.frame_locals!r},'
            u'stack={s.frame_stack!r})>'
        ).format(s=self)


class StackMapTableAttribute(Attribute):
    """
    .. note::

        Consider this experimental. This is an unnecessary 'feature' added
        in Java6 that even the official JDK has multiple bugs with. Proper
        generation of a StackMapTableAttribute requires a complete class
        hierarchy among other things.
    """
    ADDED_IN = '6.0.0'
    MINIMUM_CLASS_VERSION = (50, 0)

    def __init__(self, table, name_index=None):
        super(StackMapTableAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'StackMapTable'
            ).index
        )
        self.frames = []

    def unpack(self, info):
        # Described in "4.7.4. The StackMapTable Attribute"
        length = info.u2()
        # Start with a null-state FULL_FRAME.
        previous_frame = StackMapFrame(255)
        for i in range(length):
            frame_type = info.u1()
            frame = StackMapFrame(frame_type)
            if frame_type < 64:
                # 0 to 63 are SAME_FRAME
                if i == 0:
                    frame.frame_offset = frame_type
                else:
                    frame.frame_offset = previous_frame.frame_offset + \
                            frame_type + 1
                    frame.frame_locals = previous_frame.frame_locals

                self.frames.append(frame)
                previous_frame = frame
                continue
            elif frame_type < 128:
                # 64 to 127 are SAME_LOCALS_1_STACK_ITEM
                if i == 0:
                    frame.frame_offset = frame_type - 64
                else:
                    frame.frame_offset = previous_frame.frame_offset + \
                            frame_type - 63
                    frame.frame_locals = previous_frame.frame_locals

                frame.frame_stack = list(
                    self._unpack_verification_type_info(info, 1)
                )

                self.frames.append(frame)
                previous_frame = frame
                continue
            elif frame_type < 247:
                # Reserved types, we may be trying to parse a ClassFile that's
                # newer than we can handle.
                raise NotImplementedError()

            # All other types have an additional offset
            frame_offset = info.u2()

            if frame_type == 247:
                # SAME_LOCALS_1_STACK_ITEM_EXTENDED
                if i == 0:
                    frame.frame_offset = frame_offset
                else:
                    frame.frame_offset = previous_frame.frame_offset + \
                            frame_offset + 1
                frame.frame_locals = previous_frame.frame_locals
                frame.frame_stack = list(
                    self._unpack_verification_type_info(
                        info,
                        1
                    )
                )
            elif frame_type < 251:
                # CHOP
                if i == 0:
                    frame.frame_offset = frame_offset
                else:
                    frame.frame_offset = previous_frame.frame_offset + \
                            frame_offset + 1
                    frame.frame_locals = previous_frame.frame_locals[
                        0:251 - frame_type
                    ]
            elif frame_type == 251:
                # SAME_FRAME_EXTENDED
                if i == 0:
                    frame.frame_offset = frame_offset
                else:
                    frame.frame_offset = previous_frame.frame_offset + \
                            frame_offset + 1
                    frame.frame_locals = previous_frame.frame_locals
            elif frame_type < 255:
                # APPEND
                if i == 0:
                    frame.frame_offset = frame_offset
                else:
                    frame.frame_offset = previous_frame.frame_offset + \
                            frame_offset + 1

                frame.frame_locals = previous_frame.frame_locals + list(
                    self._unpack_verification_type_info(
                        info,
                        frame_type - 251
                    )
                )
            elif frame_type == 255:
                # FULL_FRAME
                if i == 0:
                    frame.frame_offset = frame_offset
                else:
                    frame.frame_offset = previous_frame.frame_offset + \
                            frame_offset + 1

                frame.frame_locals = list(self._unpack_verification_type_info(
                    info,
                    info.u2()
                ))
                frame.frame_stack = list(self._unpack_verification_type_info(
                    info,
                    info.u2()
                ))

            self.frames.append(frame)
            previous_frame = frame

    @staticmethod
    def _unpack_verification_type_info(info, count):
        # Unpacks the verification_type_info structure, used for both locals
        # and the stack.
        for _ in repeat(None, count):
            tag = info.u1()
            if tag in TYPES_WITH_EXTRA:
                yield (tag, info.u2())
            else:
                yield (tag,)

    def pack(self):
        raise NotImplementedError()



from struct import pack
from jawa.attribute import Attribute


class ConstantValueAttribute(Attribute):
    ADDED_IN = '1.0.2'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, value=None, name_index=None):
        super(ConstantValueAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'ConstantValue'
            ).index
        )
        self._constant_value_index = value.index if value else None

    def unpack(self, info):
        self._constant_value_index = info.u2()

    def pack(self):
        return pack('>H', self._constant_value_index)

    @property
    def constant_value(self):
        return self.cf.constants[self._constant_value_index]



import io
import inspect
import functools
from typing import Iterator
from struct import pack
from itertools import repeat
from collections import namedtuple

from jawa.attribute import Attribute, AttributeTable
from jawa.util.bytecode import (
    read_instruction,
    write_instruction,
    Instruction
)

CodeException = namedtuple('CodeException', [
    'start_pc', 'end_pc', 'handler_pc', 'catch_type'
])


class CodeAttribute(Attribute):
    """
    A `CodeAttribute` contains the executable bytecode of a single method.

    As a quick example, lets make a "HelloWorld" class with a single method
    that simply returns when it's called:

    .. code-block:: python

        from jawa import ClassFile
        from jawa.util.bytecode import Instruction

        cf = ClassFile.create('HelloWorld')

        main = cf.methods.create(
            # The name of the method
            'main',
            # The signature of the method
            '([Ljava/lang/String;)V',
            # Tell Jawa to automatically create an empty CodeAttribute for
            # us to use.
            code=True
        )
        main.code.max_locals = 1
        main.access_flags.acc_static = True
        main.code.assemble([
            Instruction.from_mnemonic('return')
        ])

        # Save it to disk so we can run it with the JVM.
        with open('HelloWorld.class', 'wb') as fout:
            cf.save(fout)
    """
    ADDED_IN = '1.0.2'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, name_index=None):
        super(CodeAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'Code'
            ).index
        )
        self.max_stack = 0
        self.max_locals = 0
        self.exception_table = []
        self.attributes = AttributeTable(table.cf, parent=self)
        self._code = ''

    def unpack(self, info):
        """
        Read the CodeAttribute from the byte string `info`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param info: A byte string containing an unparsed CodeAttribute.
        """
        self.max_stack, self.max_locals, c_len = info.unpack('>HHI')
        self._code = info.read(c_len)

        # The exception table
        ex_table_len = info.u2()
        for _ in repeat(None, ex_table_len):
            self.exception_table.append(CodeException(
                *info.unpack('>HHHH')
            ))
        self.attributes = AttributeTable(self.cf, parent=self)
        self.attributes.unpack(info)

    def pack(self):
        """
        The `CodeAttribute` in packed byte string form.
        """
        with io.BytesIO() as file_out:
            file_out.write(pack(
                '>HHI',
                self.max_stack,
                self.max_locals,
                len(self._code)
            ))
            file_out.write(self._code)

            file_out.write(pack('>H', len(self.exception_table)))
            for exception in self.exception_table:
                file_out.write(pack('>HHHH', *exception))

            self.attributes.pack(file_out)
            return file_out.getvalue()

    def assemble(self, code):
        """
        Assembles an iterable of :class:`~jawa.util.bytecode.Instruction`
        objects into a method's code body.
        """
        with io.BytesIO() as code_out:
            for ins in code:
                write_instruction(code_out, code_out.tell(), ins)
            self._code = code_out.getvalue()

    def disassemble(self, *, transforms=None) -> Iterator[Instruction]:
        """
        Disassembles this method, yielding an iterable of
        :class:`~jawa.util.bytecode.Instruction` objects.
        """
        if transforms is None:
            if self.cf.classloader:
                transforms = self.cf.classloader.bytecode_transforms
            else:
                transforms = []

        transforms = [self._bind_transform(t) for t in transforms]

        with io.BytesIO(self._code) as code:
            ins_iter = iter(lambda: read_instruction(code, code.tell()), None)
            for ins in ins_iter:
                for transform in transforms:
                    ins = transform(ins)
                yield ins

    def _bind_transform(self, transform):
        sig = inspect.signature(transform, follow_wrapped=True)
        return functools.partial(
            transform,
            **{k: v for k, v in {
                'cf': self.cf,
                'attribute': self
            }.items() if k in sig.parameters}
        )



from struct import unpack, pack

from jawa.util.utf import decode_modified_utf8, encode_modified_utf8


class Constant(object):
    """
    The base class for all ``Constant*`` types.
    """
    __slots__ = ('pool', 'index')

    def __init__(self, pool, index):
        self.pool = pool
        self.index = index


class Number(Constant):
    __slots__ = ('value',)

    def __init__(self, pool, index, value):
        super().__init__(pool, index)
        self.value = value

    def __repr__(self):
        return (
            f'{self.__class__.__name__}('
            f'index={self.index}, value={self.value!r})'
        )


class UTF8(Constant):
    __slots__ = ('value',)
    TAG = 1

    def __init__(self, pool, index, value):
        super().__init__(pool, index)
        self.value = value

    def pack(self):
        encoded_value = encode_modified_utf8(self.value)
        return pack('>BH', self.TAG, len(encoded_value)) + encoded_value

    def __repr__(self):
        return f'<UTF8(index={self.index}, value={self.value!r}>)'


class Integer(Number):
    TAG = 3

    def pack(self):
        return pack('>Bi', self.TAG, self.value)


class Float(Number):
    TAG = 4

    def pack(self):
        return pack('>Bf', self.TAG, self.value)


class Long(Number):
    TAG = 5

    def pack(self):
        return pack('>Bq', self.TAG, self.value)


class Double(Number):
    TAG = 6

    def pack(self):
        return pack('>Bd', self.TAG, self.value)


class ConstantClass(Constant):
    __slots__ = ('name_index',)
    TAG = 7

    def __init__(self, pool, index, name_index):
        super().__init__(pool, index)
        self.name_index = name_index

    @property
    def name(self):
        return self.pool.get(self.name_index)

    def pack(self):
        return pack('>BH', self.TAG, self.name_index)

    def __repr__(self):
        return f'<ConstantClass(index={self.index}, name={self.name!r})>'


class String(Constant):
    __slots__ = ('string_index',)
    TAG = 8

    def __init__(self, pool, index, string_index):
        super().__init__(pool, index)
        self.string_index = string_index

    @property
    def string(self):
        return self.pool.get(self.string_index)

    def pack(self):
        return pack('>BH', self.TAG, self.string_index)

    def __repr__(self):
        return f'<String(index={self.index}, string={self.string!r})>'


class Reference(Constant):
    __slots__ = ('class_index', 'name_and_type_index')
    TAG = None

    def __init__(self, pool, index, class_index, name_and_type_index):
        super().__init__(pool, index)
        self.class_index = class_index
        self.name_and_type_index = name_and_type_index

    @property
    def class_(self):
        return self.pool.get(self.class_index)

    @property
    def name_and_type(self):
        return self.pool.get(self.name_and_type_index)

    def pack(self):
        return pack(
            '>BHHH',
            self.TAG,
            self.class_index,
            self.name_and_type_index
        )

    def __repr__(self):
        return (
            f'<{self.__class__.__name__}('
            f'index={self.index},'
            f'class_={self.class_!r},'
            f'name_and_type={self.name_and_type!r})>'
        )


class FieldReference(Reference):
    TAG = 9


class MethodReference(Reference):
    TAG = 10


class InterfaceMethodRef(Reference):
    TAG = 11


class NameAndType(Constant):
    __slots__ = ('name_index', 'descriptor_index')
    TAG = 12

    def __init__(self, pool, index, name_index, descriptor_index):
        super().__init__(pool, index)
        self.name_index = name_index
        self.descriptor_index = descriptor_index

    @property
    def name(self):
        return self.pool.get(self.name_index)

    @property
    def descriptor(self):
        return self.pool.get(self.descriptor_index)

    def pack(self):
        return pack('>BHH', self.TAG, self.name_index, self.descriptor_index)

    def __repr__(self):
        return (
            f'<NameAndType('
            f'index={self.index},'
            f'name={self.name!r},'
            f'descriptor={self.descriptor!r})>'
        )


class MethodHandle(Constant):
    __slots__ = ('reference_kind', 'reference_index')
    TAG = 15

    def __init__(self, pool, index, reference_kind, reference_index):
        super().__init__(pool, index)
        self.reference_kind = reference_kind
        self.reference_index = reference_index

    @property
    def reference(self):
        return self.pool.get(self.reference_index)

    def pack(self):
        return pack('>BBH', self.TAG, self.reference_kind, self.reference_index)

    def __repr__(self):
        return (
            f'<MethodHandle(index={self.index}, reference={self.reference!r})>'
        )


class MethodType(Constant):
    __slots__ = ('descriptor_index',)
    TAG = 16

    def __init__(self, pool, index, descriptor_index):
        super().__init__(pool, index)
        self.descriptor_index = descriptor_index

    @property
    def descriptor(self):
        return self.pool.get(self.descriptor_index)

    def pack(self):
        return pack('>BH', self.TAG, self.descriptor_index)

    def __repr__(self):
        return f'<MethodType(index={self.index},descriptor={self.descriptor})>'


class InvokeDynamic(Constant):
    __slots__ = ('bootstrap_method_attr_index', 'name_and_type_index')
    TAG = 18

    def __init__(self, pool, index, bootstrap_method_attr_index,
                 name_and_type_index):
        super().__init__(pool, index)
        self.bootstrap_method_attr_index = bootstrap_method_attr_index
        self.name_and_type_index = name_and_type_index

    @property
    def method_attr_index(self):
        return self.bootstrap_method_attr_index

    @property
    def name_and_type(self):
        return self.pool[self.name_and_type_index]

    def pack(self):
        return pack(
            '>BHH',
            self.TAG,
            self.bootstrap_method_attr_index,
            self.name_and_type_index
        )

    def __repr__(self):
        return (
            f'<InvokeDynamic('
            f'index={self.index},'
            f'method_attr_index={self.method_attr_index},'
            f'name_and_type={self.name_and_type!r})>'
        )


class Module(ConstantClass):
    __slots__ = ('name_index',)
    TAG = 19

    def __repr__(self):
        return f'<Module(index={self.index}, name={self.name!r})>'


class PackageInfo(ConstantClass):
    __slots__ = ('name_index',)
    TAG = 20

    def __repr__(self):
        return f'<PackageInfo(index={self.index}, name={self.name!r})>'


_constant_types = (
    None,
    UTF8,
    None,
    Integer,
    Float,
    Long,
    Double,
    ConstantClass,
    String,
    FieldReference,
    MethodReference,
    InterfaceMethodRef,
    NameAndType,
    None,
    None,
    MethodHandle,
    MethodType,
    None,
    InvokeDynamic,
    Module,
    PackageInfo
)


# The format and size-on-disk of each type of constant
# in the constant pool.
_constant_fmts = (
    None, None, None,
    ('>i', 4),
    ('>f', 4),
    ('>q', 8),
    ('>d', 8),
    ('>H', 2),
    ('>H', 2),
    ('>HH', 4),
    ('>HH', 4),
    ('>HH', 4),
    ('>HH', 4),
    None,
    None,
    ('>BH', 3),
    ('>H', 2),
    None,
    ('>HH', 4)
)


class ConstantPool(object):
    def __init__(self):
        self._pool = [None]

    def append(self, constant):
        """
        Appends a new constant to the end of the pool.
        """
        self._pool.append(constant)

    def __iter__(self):
        for index, constant in enumerate(self._pool):
            if constant is not None:
                yield self.get(index)

    def get(self, index):
        """
        Returns the `Constant` at `index`, raising a KeyError if it
        does not exist.
        """
        constant = self._pool[index]
        if not isinstance(constant, Constant):
            constant = _constant_types[constant[0]](self, index, *constant[1:])
            self._pool[index] = constant
        return constant

    def __getitem__(self, idx):
        return self.get(idx)

    def __setitem__(self, idx, value):
        self._pool[idx] = value

    def find(self, type_=None, f=None):
        """
        Iterates over the pool, yielding each matching ``Constant``. Calling
        without any arguments is equivalent to iterating over the pool.

        :param type_: Any subclass of :class:`Constant` or ``None``.
        :param f: Any callable which takes one argument (the constant).
        """
        for constant in self:
            if type_ is not None and not isinstance(constant, type_):
                continue

            if f is not None and not f(constant):
                continue

            yield constant

    def find_one(self, *args, **kwargs):
        """
        Same as ``find()`` but returns only the first result, or `None` if
        nothing was found.
        """
        try:
            return next(self.find(*args, **kwargs))
        except StopIteration:
            return None

    def create_utf8(self, value):
        """
        Creates a new :class:`ConstantUTF8`, adding it to the pool and
        returning it.

        :param value: The value of the new UTF8 string.
        """
        self.append((1, value))
        return self.get(self.raw_count - 1)

    def create_integer(self, value: int) -> Integer:
        """
        Creates a new :class:`ConstantInteger`, adding it to the pool and
        returning it.

        :param value: The value of the new integer.
        """
        self.append((3, value))
        return self.get(self.raw_count - 1)

    def create_float(self, value: float) -> Float:
        """
        Creates a new :class:`ConstantFloat`, adding it to the pool and
        returning it.

        :param value: The value of the new float.
        """
        self.append((4, value))
        return self.get(self.raw_count - 1)

    def create_long(self, value: int) -> Long:
        """
        Creates a new :class:`ConstantLong`, adding it to the pool and
        returning it.

        :param value: The value of the new long.
        """
        self.append((5, value))
        self.append(None)
        return self.get(self.raw_count - 2)

    def create_double(self, value: float) -> Double:
        """
        Creates a new :class:`ConstantDouble`, adding it to the pool and
        returning it.

        :param value: The value of the new Double.
        """
        self.append((6, value))
        self.append(None)
        return self.get(self.raw_count - 2)

    def create_class(self, name: str) -> ConstantClass:
        """
        Creates a new :class:`ConstantClass`, adding it to the pool and
        returning it.

        :param name: The name of the new class.
        """
        self.append((
            7,
            self.create_utf8(name).index
        ))
        return self.get(self.raw_count - 1)

    def create_string(self, value: str) -> String:
        """
        Creates a new :class:`ConstantString`, adding it to the pool and
        returning it.

        :param value: The value of the new string as a UTF8 string.
        """
        self.append((
            8,
            self.create_utf8(value).index
        ))
        return self.get(self.raw_count - 1)

    def create_name_and_type(self, name: str, descriptor: str) -> NameAndType:
        """
        Creates a new :class:`ConstantNameAndType`, adding it to the pool and
        returning it.

        :param name: The name of the class.
        :param descriptor: The descriptor for `name`.
        """
        self.append((
            12,
            self.create_utf8(name).index,
            self.create_utf8(descriptor).index
        ))
        return self.get(self.raw_count - 1)

    def create_field_ref(self, class_: str, field: str, descriptor: str) \
            -> FieldReference:
        """
        Creates a new :class:`ConstantFieldRef`, adding it to the pool and
        returning it.

        :param class_: The name of the class to which `field` belongs.
        :param field: The name of the field.
        :param descriptor: The descriptor for `field`.
        """
        self.append((
            9,
            self.create_class(class_).index,
            self.create_name_and_type(field, descriptor).index
        ))
        return self.get(self.raw_count - 1)

    def create_method_ref(self, class_: str, method: str, descriptor: str) \
            -> MethodReference:
        """
        Creates a new :class:`ConstantMethodRef`, adding it to the pool and
        returning it.

        :param class_: The name of the class to which `method` belongs.
        :param method: The name of the method.
        :param descriptor: The descriptor for `method`.
        """
        self.append((
            10,
            self.create_class(class_).index,
            self.create_name_and_type(method, descriptor).index
        ))
        return self.get(self.raw_count - 1)

    def create_interface_method_ref(self, class_: str, if_method: str,
                                    descriptor: str) -> InterfaceMethodRef:
        """
        Creates a new :class:`ConstantInterfaceMethodRef`, adding it to the
        pool and returning it.

        :param class_: The name of the class to which `if_method` belongs.
        :param if_method: The name of the interface method.
        :param descriptor: The descriptor for `if_method`.
        """
        self.append((
            11,
            self.create_class(class_).index,
            self.create_name_and_type(if_method, descriptor).index
        ))
        return self.get(self.raw_count - 1)

    def unpack(self, fio):
        """
        Read the ConstantPool from the file-like object `fio`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param fio: Any file-like object providing `read()`
        """
        # Reads in the ConstantPool (constant_pool in the JVM Spec)
        constant_pool_count = unpack('>H', fio.read(2))[0]

        # Pull this locally so CPython doesn't do a lookup each time.
        read = fio.read

        while constant_pool_count > 1:
            constant_pool_count -= 1
            # The 1-byte prefix identifies the type of constant.
            tag = ord(read(1))

            if tag == 1:
                # CONSTANT_Utf8_info, a length prefixed UTF-8-ish string.
                # Only attempt to properly decode the MUTF8 if it fails
                # regular UTF8 decoding, which overs huge time savings over
                # large JARs.
                utf8_str = read(unpack('>H', read(2))[0])
                try:
                    utf8_str = utf8_str.decode('utf8')
                except UnicodeDecodeError:
                    utf8_str = decode_modified_utf8(utf8_str)
                self.append((tag, utf8_str))
            else:
                # Every other constant type is trivial.
                fmt, size = _constant_fmts[tag]
                self.append((tag, *unpack(fmt, read(size))))
                if tag == 5 or tag == 6:
                    # LONG (5) and DOUBLE (6) count as two entries in the
                    # pool.
                    self.append(None)
                    constant_pool_count -= 1

    def pack(self, fout):
        """
        Write the ConstantPool to the file-like object `fout`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be calle=d for you when saving a ClassFile.

        :param fout: Any file-like object providing `write()`
        """
        write = fout.write
        write(pack('>H', self.raw_count))

        for constant in self:
            write(constant.pack())

    def __len__(self) -> int:
        """
        The number of `Constants` in the `ConstantPool`, excluding padding.
        """
        count = 0
        for constant in self._pool:
            if constant is not None:
                count += 1
        return count

    @property
    def raw_count(self) -> int:
        """
        The number of `Constants` in the `ConstantPool`, including padding.
        """
        return len(self._pool)



from struct import pack
from collections import namedtuple

from jawa.attribute import Attribute


local_variable_entry = namedtuple('local_variable_entry', [
    'start_pc',
    'length',
    'name_index',
    'descriptor_index',
    'index'
])


class LocalVariableTableAttribute(Attribute):
    ADDED_IN = '1.0.2'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, name_index=None):
        super().__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'LocalVariableTable'
            ).index
        )
        self.local_variables = []

    def unpack(self, info):
        length = info.u2()
        table = info.unpack('>{0}H'.format(length * 5))

        self.local_variables = [
            local_variable_entry(*x)
            for x in zip(*[iter(table)] * 5)
        ]

    def pack(self):
        return pack(
            '>H{0}H'.format(len(self.local_variables) * 5),
            len(self.local_variables),
            *sum(self.local_variables, ())
        )

    def __repr__(self):
        return f'<LocalVariableTableAttribute({self.local_variables!r})>'





class VerificationTypes(object):
    ITEM_Top = 0
    ITEM_Integer = 1
    ITEM_Float = 2
    ITEM_Long = 4
    ITEM_Double = 3
    ITEM_Null = 5
    ITEM_UninitializedThis = 6
    ITEM_Object = 7
    ITEM_Uninitialized = 8



from struct import pack
from jawa.attribute import Attribute


class SignatureAttribute(Attribute):
    ADDED_IN = '5.0.0'
    MINIMUM_CLASS_VERSION = (49, 0)

    def __init__(self, table, name_index):
        super(SignatureAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'Signature'
            ).index
        )
        self._signature_index = None

    def unpack(self, info):
        self._signature_index = info.u2()

    def pack(self):
        return pack('>H', self._signature_index)

    @property
    def signature(self):
        return self.cf.constants[self._signature_index]

    @signature.setter
    def signature(self, value):
        self._signature_index = value.index



import inspect
import pkgutil
import importlib
from typing import IO, Callable, Iterator, Union, Dict, Any, Tuple
from struct import unpack, pack
from itertools import repeat

from jawa.constants import UTF8
from jawa.util.stream import BufferStreamReader


class Attribute(object):
    ADDED_IN: int = None
    MINIMUM_CLASS_VERSION: Tuple[int, int] = None

    def __init__(self, parent: 'AttributeTable', name_index: int):
        self.parent = parent
        self.name_index = name_index

    @property
    def name(self) -> UTF8:
        """
        The name of this attribute.
        """
        return self.cf.constants[self.name_index]

    @property
    def cf(self):
        """
        The ClassFile that owns this attribute, if any.
        """
        return self.parent.cf

    def unpack(self, info: Union[bytes, BufferStreamReader]):
        """
        Parses an instance of this attribute from the blob `info`.
        """
        raise NotImplementedError()

    def pack(self) -> bytes:
        """
        This attribute packed into its on-disk representation.
        """
        raise NotImplementedError()


class UnknownAttribute(Attribute):
    def __init__(self, parent: 'AttributeTable', name_index: int):
        super().__init__(parent, name_index)
        self.info = None

    def unpack(self, info: Union[bytes, BufferStreamReader]):
        self.info = info

    def pack(self) -> bytes:
        return self.info


class AttributeTable(object):
    def __init__(self, cf, parent: Attribute=None):
        #: The ClassFile that ultimately owns this AttributeTable.
        self.cf = cf
        #: The parent Attribute, if one exists.
        self.parent = parent
        self._table = []

    def unpack(self, source: IO):
        """
        Read the ConstantPool from the file-like object `source`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when loading a ClassFile.

        :param source: Any file-like object providing `read()`
        """
        count = unpack('>H', source.read(2))[0]
        for _ in repeat(None, count):
            name_index, length = unpack('>HI', source.read(6))
            info_blob = source.read(length)
            self._table.append((name_index, info_blob))

    def __getitem__(self, key):
        attr = self._table[key]

        if not isinstance(attr, Attribute):
            name_index, info = attr[0], attr[1]
            name = self.cf.constants[name_index].value

            attribute_type = ATTRIBUTE_CLASSES.get(name, UnknownAttribute)
            self._table[key] = attr = attribute_type(self, name_index)
            if attribute_type is UnknownAttribute:
                attr.unpack(info)
            else:
                attr.unpack(BufferStreamReader(info))

        return attr

    def __len__(self):
        return len(self._table)

    def pack(self, out: IO):
        """
        Write the AttributeTable to the file-like object `out`.

        .. note::

            Advanced usage only. You will typically never need to call this
            method as it will be called for you when saving a ClassFile.

        :param out: Any file-like object providing `write()`
        """
        out.write(pack('>H', len(self._table)))
        for attribute in self:
            info = attribute.pack()
            out.write(pack(
                '>HI',
                attribute.name.index,
                len(info)
            ))
            out.write(info)

    def create(self, type_, *args, **kwargs) -> Any:
        """
        Creates a new attribute of `type_`, appending it to the attribute
        table and returning it.
        """
        attribute = type_(self, *args, **kwargs)
        self._table.append(attribute)
        return attribute

    def find(self, *, name: str=None, f: Callable=None) -> Iterator[Any]:
        for idx, attribute in enumerate(self._table):
            if name is not None:
                # Optimization to filter solely on name without causing
                # a full attribute load.
                if not isinstance(attribute, Attribute) and f is None:
                    attr_name = self.cf.constants[attribute[0]].value
                    if attr_name != name:
                        continue
                elif name != attribute.name.value:
                    continue

            # Force an attribute load.
            if not isinstance(attribute, Attribute):
                attribute = self[idx]

            if f is not None and not f(attribute):
                continue

            yield attribute

    def find_one(self, **kwargs) -> Any:
        """
        Same as ``find()`` but returns only the first result.
        """
        return next(self.find(**kwargs), None)


def get_attribute_classes() -> Dict[str, Attribute]:
    """
    Lookup all builtin Attribute subclasses, load them, and return a dict
    """
    attribute_children = pkgutil.iter_modules(
        importlib.import_module('jawa.attributes').__path__,
        prefix='jawa.attributes.'
    )

    result = {}
    for _, name, _ in attribute_children:
        classes = inspect.getmembers(
            importlib.import_module(name),
            lambda c: (
                inspect.isclass(c) and issubclass(c, Attribute) and
                c is not Attribute
            )
        )

        for class_name, class_ in classes:
            attribute_name = getattr(class_, 'ATTRIBUTE_NAME', class_name[:-9])
            result[attribute_name] = class_

    return result


#: A dictionary of known attribute subclasses at the time this module
#: was loaded.
ATTRIBUTE_CLASSES = get_attribute_classes()



import io
from struct import pack
from itertools import repeat
from collections import namedtuple
from jawa.attribute import Attribute


InnerClass = namedtuple('InnerClass', [
    'inner_class_info_index',
    'outer_class_info_index',
    'inner_name_index',
    'inner_class_access_flags'
])


class InnerClassesAttribute(Attribute):
    ADDED_IN = '1.1.0'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, name_index):
        super().__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'InnerClasses'
            ).index
        )
        self.inner_classes = []

    def unpack(self, info):
        for _ in repeat(None, info.u2()):
            self.inner_classes.append(InnerClass(*info.unpack('>HHHH')))

    def pack(self):
        with io.BytesIO() as out:
            out.write(pack('>H', len(self.inner_classes)))
            for inner_class in self.inner_classes:
                out.write(pack('>HHHH', *inner_class))
            return out.getvalue()




from setuptools import setup, find_packages

setup(
    name="jawa",
    packages=find_packages(),
    version="2.0",
    description="Doing fun stuff with JVM ClassFiles.",
    author="Tyler Kennedy",
    author_email="tk@tkte.ch",
    url="http://github.com/TkTech/Jawa",
    keywords=["java", "disassembly", "disassembler"],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Disassemblers"
    ],
    install_requires=[
        'click>=5.0'
    ],
    tests_require=[
        'pytest>=2.10',
    ],
    extras_require={
        'dev': [
            'pytest',
            'sphinx',
            'sphinxcontrib-googleanalytics',
            'sphinx_rtd_theme',
            'ghp-import',
            'pyyaml',
            'ipython'
        ]
    },
    entry_points='''
    [console_scripts]
    jawa=jawa.cli:cli
    '''
)



from jawa.attribute import Attribute


class SyntheticAttribute(Attribute):
    ADDED_IN = '5.0.0'
    MINIMUM_CLASS_VERSION = (49, 0)

    def __init__(self, table, name_index=None):
        super().__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'Synthetic'
            ).index
        )

    def pack(self):
        pass

    def unpack(self, info):
        pass



import re
import json

import click

from jawa.classloader import ClassLoader
from jawa.cf import ClassVersion, ClassFile
from jawa.attribute import get_attribute_classes
from jawa.util import bytecode, shell
from jawa.constants import UTF8


@click.group()
def cli():
    pass


@cli.command()
def attributes():
    """List enabled Attributes.

    Prints a list of all enabled ClassFile Attributes.
    """
    attribute_classes = get_attribute_classes()
    for name, class_ in attribute_classes.items():
        click.echo(
            u'{name} - Added in: {ai} ({cv})'.format(
                name=click.style(name, fg='green'),
                ai=click.style(class_.ADDED_IN, fg='yellow'),
                cv=click.style(
                    ClassVersion(*class_.MINIMUM_CLASS_VERSION).human,
                    fg='yellow'
                )
            )
        )


@cli.command()
@click.argument('mnemonic')
def ins(mnemonic):
    """Lookup instruction information.

    Lookup an instruction by its mnemonic.
    """
    try:
        opcode = bytecode.opcode_table[mnemonic]
    except KeyError:
        click.secho(u'No definition found.', fg='red')
        return

    click.echo(u'{mnemonic} (0x{op})'.format(
        mnemonic=click.style(opcode['mnemonic'], fg='green', underline=True),
        op=click.style(format(opcode['op'], '02x'), fg='green')
    ))

    if opcode.get('desc'):
        click.secho('Description:', fg='yellow')
        click.echo(opcode['desc'])

    if opcode['can_be_wide']:
        click.echo(u'This instruction can be prefixed by the WIDE opcode.')

    if opcode.get('runtime'):
        click.secho('Possible runtime exceptions:', fg='yellow')
        for runtime_exception in opcode['runtime']:
            click.echo('- {runtime_exception}'.format(
                runtime_exception=click.style(runtime_exception, fg='red')
            ))

    if opcode['operands']:
        click.secho(u'Operand Format:', fg='yellow')
        for operand_fmt, operand_type in opcode['operands']:
            click.echo(u'- {ty} as a {fmt}'.format(
                ty=click.style(operand_type.name, fg='yellow'),
                fmt=click.style(operand_fmt.name, fg='yellow')
            ))
    elif opcode['op'] in (0xAB, 0xAA, 0xC4):
        # lookup[table|switch] and WIDE.
        click.secho(u'\nOperand Format:', fg='yellow')
        click.echo(
            u'This is a special-case opcode with variable operand parsing.'
        )


@cli.command(name='shell')
@click.option('--class-path', '-cp', multiple=True)
def shell_command(class_path):
    """Drop into a debugging shell.

    Once the shell is loaded you can use `load(<class name>)` to load
    any class on the set classpath.
    """
    loader = ClassLoader(*class_path)
    shell.start_shell(local_ns={
        'ClassFile': ClassFile,
        'load': loader.load,
        'loader': loader
    })


@cli.command(name='def2json')
@click.argument('source', type=click.File('rb'))
def definition_to_json(source):
    """Convert a bytecode.yaml file into a prepared bytecode.json.

    Jawa internally uses a YAML file to define all bytecode opcodes, operands,
    runtime exceptions, default transforms, etc...

    However since JSON is available in the python stdlib and YAML is not, we
    process this YAML file before distribution to prevent adding an unnecessary
    dependency.
    """
    try:
        import yaml
    except ImportError:
        click.echo(
            'The pyyaml module could not be found and is required'
            ' to use this command.',
            err=True
        )
        return

    y = yaml.load(source)

    for k, v in y.items():
        # We guarantee some keys should always exist to make life easier for
        # developers.
        v.setdefault('operands', None)
        v.setdefault('can_be_wide', False)
        v.setdefault('transform', {})
        v['mnemonic'] = k

    click.echo(json.dumps(y, indent=4, sort_keys=True))


@cli.command()
@click.argument('source', type=click.Path(exists=True))
def dependencies(source):
    """Output a list of all classes referenced by the given source."""
    loader = ClassLoader(source, max_cache=-1)
    all_dependencies = set()
    for klass in loader.classes:
        new_dependencies = loader.dependencies(klass) - all_dependencies
        all_dependencies.update(new_dependencies)
        for new_dep in new_dependencies:
            click.echo(new_dep)


@cli.command()
@click.argument('source', type=click.Path(exists=True))
@click.argument('regex')
@click.option(
    '--stop-on-first',
    default=False,
    is_flag=True,
    help='Stop iteration on first matching class.'
)
def grep(source, regex, stop_on_first=False):
    """Grep the constant pool of all classes in source."""
    loader = ClassLoader(source, max_cache=-1)
    r = re.compile(regex)

    def _matches(constant):
        return r.match(constant.value)

    for klass in loader.classes:
        it = loader.search_constant_pool(path=klass, type_=UTF8, f=_matches)
        if next(it, None):
            print(klass)
            if stop_on_first:
                break



import io
from collections import namedtuple
from itertools import repeat
from struct import pack

from jawa.attribute import Attribute

BootstrapMethod = namedtuple(
    'BootstrapMethod',
    ['method_ref', 'bootstrap_args']
)


class BootstrapMethodsAttribute(Attribute):
    ADDED_IN = '7'
    MINIMUM_CLASS_VERSION = (51, 0)

    def __init__(self, table, name_index=None):
        super(BootstrapMethodsAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'BootstrapMethods'
            ).index
        )
        self.table = []

    def __repr__(self):
        return '<BootstrapMethods(table={self.table})>'.format(
            self=self
        )

    def pack(self):
        out = io.BytesIO()
        out.write(pack('>H', len(self.table)))

        for table_entry in self.table:
            out.write(pack(
                '>HH',
                table_entry.method_ref,
                len(table_entry.bootstrap_args)
            ))
            out.write(pack(
                '>{0}H'.format(len(table_entry.bootstrap_args)),
                table_entry.bootstrap_args
            ))

        return out.getvalue()

    def unpack(self, info):
        length = info.u2()

        for _ in repeat(None, length):
            self.table.append(BootstrapMethod(
                info.u2(),
                info.unpack('>{0}H'.format(info.u2()))
            ))



from jawa.attribute import Attribute


class DeprecatedAttribute(Attribute):
    ADDED_IN = '1.1.0'
    MINIMUM_CLASS_VERSION = (45, 3)

    def __init__(self, table, name_index=None):
        super(DeprecatedAttribute, self).__init__(
            table,
            name_index or table.cf.constants.create_utf8(
                'Deprecated'
            ).index
        )

    def __repr__(self):
        return '<DeprecatedAttribute()>'

    def pack(self):
        pass

    def unpack(self, info):
        pass
